import markdown

html = markdown.markdownFromFile(
        input="documentation_fr.md",
        extensions=[
            'markdown.extensions.codehilite',
            'markdown.extensions.toc',
            'markdown.extensions.tables'
        ],
        extension_configs={
            'markdown.extensions.codehilite': {
                'noclasses': True,
            },
            'markdown.extensions.toc': {
                'title': 'Sommaire'
            }
        })



def run(action, key_start_time=0, other_only=False, **kwargs):
    if other_only:
        action.mapping.stop_all_running(
                except_key=action.key,
                key_start_time=key_start_time)
    else:
        action.mapping.stop_all_running()

def description(action, other_only=False, **kwargs):
    message = "stopping all actions"
    if other_only:
        message += " except this key"

    return _(message)



from transitions.extensions import HierarchicalMachine as Machine
from .helpers import debug_print, error_print
from . import actions

class Action:
    STATES = [
        'initial',
        'loading',
        'failed',
        {
            'name': 'loaded',
            'children': ['running']
        }
    ]

    TRANSITIONS = [
        {
            'trigger': 'load',
            'source': 'initial',
            'dest': 'loading'
        },
        {
            'trigger': 'fail',
            'source': 'loading',
            'dest': 'failed',
            'after': 'poll_loaded'
        },
        {
            'trigger': 'success',
            'source': 'loading',
            'dest': 'loaded',
            'after': 'poll_loaded'
        },
        {
            'trigger': 'run',
            'source': 'loaded',
            'dest': 'loaded_running',
            'after': 'finish_action',
            # if a child has no transitions, then it is bubbled to the parent,
            # and we don't want that. Not useful in that machine precisely.
            'conditions': ['is_loaded']
        },
        {
            'trigger': 'finish_action',
            'source': 'loaded_running',
            'dest': 'loaded'
        }
    ]

    def __init__(self, action, key, **kwargs):
        Machine(model=self, states=self.STATES,
                transitions=self.TRANSITIONS, initial='initial',
                ignore_invalid_triggers=True, queued=True)

        self.action = action
        self.key = key
        self.mapping = key.parent
        self.arguments = kwargs
        self.sleep_event = None
        self.waiting_music = None

    def is_loaded_or_failed(self):
        return self.is_loaded(allow_substates=True) or self.is_failed()

    def callback_music_loaded(self, success):
        if success:
            self.success()
        else:
            self.fail()

    # Machine states / events
    def on_enter_loading(self):
        if hasattr(actions, self.action):
            if 'music' in self.arguments:
                self.arguments['music'].subscribe_loaded(
                        self.callback_music_loaded)
            else:
                self.success()
        else:
            error_print("Unknown action {}".format(self.action))
            self.fail()

    def on_enter_loaded_running(self, key_start_time):
        debug_print(self.description())
        if hasattr(actions, self.action):
            getattr(actions, self.action).run(self,
                    key_start_time=key_start_time, **self.arguments)

    def poll_loaded(self):
        self.key.callback_action_ready(self,
                self.is_loaded(allow_substates=True))

    # This one cannot be in the Machine state since it would be queued to run
    # *after* the wait is ended...
    def interrupt(self):
        if getattr(actions, self.action, None) and\
                hasattr(getattr(actions, self.action), 'interrupt'):
            return getattr(getattr(actions, self.action), 'interrupt')(
                    self, **self.arguments)

    # Helpers
    def music_list(self, music):
        if music is not None:
            return [music]
        else:
            return self.mapping.open_files.values()

    def description(self):
        if hasattr(actions, self.action):
            return getattr(actions, self.action)\
                    .description(self, **self.arguments)
        else:
            return _("unknown action {}").format(self.action)



from kivy.uix.label import Label
from kivy.uix.stacklayout import StackLayout
from kivy.uix.relativelayout import RelativeLayout
from kivy.properties import ListProperty
from kivy.clock import Clock
from ..helpers import duration_to_min_sec
from ..lock import Lock

__all__ = ["PlayList",
        "PlayListIcons", "PlayListIcon",
        "PlayListNames", "PlayListName",
        "PlayListTimes", "PlayListTime"]

playlist_lock = Lock("playlist")

class PlayList(RelativeLayout):
    playlist = ListProperty([])

    def __init__(self, **kwargs):
        super(PlayList, self).__init__(**kwargs)
        Clock.schedule_interval(self.update_playlist, 0.5)

    def update_playlist(self, dt):
        if self.parent is None or 'Mapping' not in self.parent.ids:
            return True

        open_files = self.parent.ids['Mapping'].open_files
        playlist = []
        for music_file in open_files.values():
            if not music_file.is_in_use():
                continue

            time_info = "{}/{}".format(
                    duration_to_min_sec(music_file.sound_position),
                    duration_to_min_sec(music_file.sound_duration))

            if music_file.is_loaded_paused():
                playlist.append(["⏸", music_file.name, time_info, False])
            else:
                playlist.append(["⏵", music_file.name, time_info, True])
        with playlist_lock:
            self.playlist = playlist


class PlayListIcons(StackLayout):
    def __init__(self, **kwargs):
        super(PlayListIcons, self).__init__(**kwargs)
        self.icons = []

    def on_parent(self, instance, parent):
        parent.bind(playlist=self.update_playlist_icons)

    def update_playlist_icons(self, instance, playlist):
        icons_length = len(self.icons)
        index = -1
        for index, [icon, filename, time_info, playing] in enumerate(playlist):
            if index >= icons_length:
                icon_label = PlayListIcon(text=icon)
                self.add_widget(icon_label)
                self.icons.append(icon_label)
            else:
                self.icons[index].text = icon

        if index+1 < icons_length:
            self.clear_widgets(children=self.icons[index+1:icons_length])
            del(self.icons[index+1:icons_length])

class PlayListIcon(Label):
    def __init__(self, text='', **kwargs):
        super(PlayListIcon, self).__init__(**kwargs)
        self.text = text

    def on_parent(self, instance, parent):
        if parent is not None:
            parent.bind(font_size=self.update_font_size)
            parent.bind(labels_height=self.update_height)

    def update_height(self, instance, height):
        self.height = height

    def update_font_size(self, instance, font_size):
        self.font_size = font_size

class PlayListNames(StackLayout):
    def __init__(self, **kwargs):
        super(PlayListNames, self).__init__(**kwargs)
        self.names = []

    def on_parent(self, instance, parent):
        parent.bind(playlist=self.update_playlist_names)

    def update_playlist_names(self, instance, playlist):
        names_length = len(self.names)
        index = -1
        for index, [icon, filename, time_info, playing] in enumerate(playlist):
            if index >= names_length:
                name_label = PlayListName(text=filename, is_playing=playing)
                self.add_widget(name_label)
                self.names.append(name_label)
            else:
                self.names[index].text = filename

        if index+1 < names_length:
            self.clear_widgets(children=self.names[index+1:names_length])
            del(self.names[index+1:names_length])

class PlayListName(Label):
    def __init__(self, text='', is_playing=False, **kwargs):
        super(PlayListName, self).__init__(**kwargs)
        self.text = text
        self.bold = is_playing

    def on_parent(self, instance, parent):
        if parent is not None:
            parent.bind(font_size=self.update_font_size)
            parent.bind(labels_height=self.update_height)

    def update_height(self, instance, height):
        self.height = height

    def update_font_size(self, instance, font_size):
        self.font_size = font_size

class PlayListTimes(StackLayout):
    def __init__(self, **kwargs):
        super(PlayListTimes, self).__init__(**kwargs)
        self.times = []

    def on_parent(self, instance, parent):
        parent.bind(playlist=self.update_playlist_times)

    def update_playlist_times(self, instance, playlist):
        times_length = len(self.times)
        index = -1
        for index, [icon, filename, time_info, playing] in enumerate(playlist):
            if index >= times_length:
                time_label = PlayListTime(text=time_info)
                self.add_widget(time_label)
                self.times.append(time_label)
            else:
                self.times[index].text = time_info

        if index+1 < times_length:
            self.clear_widgets(children=self.times[index+1:times_length])
            del(self.times[index+1:times_length])

class PlayListTime(Label):
    def __init__(self, text='', **kwargs):
        super(PlayListTime, self).__init__(**kwargs)
        self.text = text

    def on_parent(self, instance, parent):
        if parent is not None:
            parent.bind(font_size=self.update_font_size)
            parent.bind(labels_height=self.update_height)

    def update_height(self, instance, height):
        self.height = height

    def update_font_size(self, instance, font_size):
        self.font_size = font_size




def run(action, music=None, value=0, delta=False, **kwargs):
    for music in action.music_list(music):
        music.seek(value=value, delta=delta)

def description(action, music=None, value=0, delta=False, **kwargs):
    if delta:
        if music is not None:
            return _("moving music « {} » by {:+d}s") \
                    .format(music.name, value)
        else:
            return _("moving all musics by {:+d}s") \
                    .format(value)
    else:
        if music is not None:
            return _("moving music « {} » to position {}s") \
                    .format(music.name, value)
        else:
            return _("moving all musics to position {}s") \
                    .format(value)



# This file was imported from
# https://bitbucket.org/marcusva/python-utils/overview
# And slightly adapted

"""OS-specific font detection."""
import os
import sys
from subprocess import Popen, PIPE

if sys.platform in ("win32", "cli"):
    import winreg

__all__ = ["STYLE_NORMAL", "STYLE_BOLD", "STYLE_ITALIC", "STYLE_LIGHT",
           "init", "list_fonts", "get_fonts", "get_font"
           ]

# Font cache entries:
# { family : [...,
#             (name, styles, fonttype, filename)
#             ...
#            ]
# }
__FONTCACHE = None


STYLE_NONE   = 0x00
STYLE_NORMAL = 0x01
STYLE_BOLD   = 0x02
STYLE_ITALIC = 0x04
STYLE_LIGHT  = 0x08
STYLE_MEDIUM = 0x10

def _add_font(family, name, styles, fonttype, filename):
    """Adds a font to the internal font cache."""
    global __FONTCACHE

    if family not in __FONTCACHE:
        __FONTCACHE[family] = []
    __FONTCACHE[family].append((name, styles, fonttype, filename))


def _cache_fonts_win32():
    """Caches fonts on a Win32 platform."""
    key = "SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\Fonts"
    regfonts = []
    try:
        with winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, key) as fontkey:
            idx = 0
            enumval = winreg.EnumValue
            rappend = regfonts.append
            while True:
                rappend(enumval(fontkey, idx)[:2])
                idx += 1
    except WindowsError:
        pass

    # TODO: integrate alias handling for fonts within the registry.
    # TODO: Scan and index fonts from %SystemRoot%\\Fonts that are not in the
    # registry

    # Received all fonts from the registry.
    for name, filename in regfonts:
        fonttype = os.path.splitext(filename)[1][1:].lower()
        if name.endswith("(TrueType)"):
            name = name[:-10].strip()
        if name.endswith("(All Res)"):
            name = name[:-9].strip()
        style = STYLE_NORMAL
        if name.find(" Bold") >= 0:
            style |= STYLE_BOLD
        if name.find(" Italic") >= 0 or name.find(" Oblique") >= 0:
            style |= STYLE_ITALIC

        family = name
        for rm in ("Bold", "Italic", "Oblique"):
            family = family.replace(rm, "")
        family = family.lower().strip()

        fontpath = os.environ.get("SystemRoot", "C:\\Windows")
        fontpath = os.path.join(fontpath, "Fonts")
        if filename.find("\\") == -1:
            # No path delimiter is given; we assume it to be a font in
            # %SystemRoot%\Fonts
            filename = os.path.join(fontpath, filename)
        _add_font(family, name, style, fonttype, filename)


def _cache_fonts_darwin():
    """Caches fonts on Mac OS."""
    raise NotImplementedError("Mac OS X support is not given yet")


def _cache_fonts_fontconfig():
    """Caches font on POSIX-alike platforms."""
    try:
        command = "fc-list : file family style fullname fullnamelang"
        proc = Popen(command, stdout=PIPE, shell=True, stderr=PIPE)
        pout = proc.communicate()[0]
        output = pout.decode("utf-8")
    except OSError:
        return

    for entry in output.split(os.linesep):
        if entry.strip() == "":
            continue
        values = entry.split(":")
        filename = values[0]

        # get the font type
        fname, fonttype = os.path.splitext(filename)
        if fonttype == ".gz":
            fonttype = os.path.splitext(fname)[1][1:].lower()
        else:
            fonttype = fonttype.lstrip(".").lower()

        # get the font name
        name = None
        if len(values) > 3:
            fullnames, fullnamelangs = values[3:]
            langs = fullnamelangs.split(",")
            try:
                offset = langs.index("fullnamelang=en")
            except ValueError:
                offset = -1
            if offset == -1:
                try:
                    offset = langs.index("en")
                except ValueError:
                    offset = -1
            if offset != -1:
                # got an english name, use that one
                name = fullnames.split(",")[offset]
                if name.startswith("fullname="):
                    name = name[9:]
        if name is None:
            if fname.endswith(".pcf") or fname.endswith(".bdf"):
                name = os.path.basename(fname[:-4])
            else:
                name = os.path.basename(fname)
        name = name.lower()

        # family and styles
        family = values[1].strip().lower()
        stylevals = values[2].strip()
        style = STYLE_NONE

        if stylevals.find("Bold") >= 0:
            style |= STYLE_BOLD
        if stylevals.find("Light") >= 0:
            style |= STYLE_LIGHT
        if stylevals.find("Italic") >= 0 or stylevals.find("Oblique") >= 0:
            style |= STYLE_ITALIC
        if stylevals.find("Medium") >= 0:
            style |= STYLE_MEDIUM
        if style == STYLE_NONE:
            style = STYLE_NORMAL
        _add_font(family, name, style, fonttype, filename)


def init():
    """Initialises the internal font cache.

    It does not need to be called explicitly.
    """
    global __FONTCACHE
    if __FONTCACHE is not None:
        return
    __FONTCACHE = {}
    if sys.platform in ("win32", "cli"):
        _cache_fonts_win32()
    elif sys.platform == "darwin":
        _cache_fonts_darwin()
    else:
        _cache_fonts_fontconfig()


def list_fonts():
    """Returns an iterator over the cached fonts."""
    if __FONTCACHE is None:
        init()
    if len(__FONTCACHE) == 0:
        yield None
    for family, entries in __FONTCACHE.items():
        for fname, styles, fonttype, filename in entries:
            yield (family, fname, styles, fonttype, filename)


def get_fonts(name, style=STYLE_NONE, ftype=None):
    """Retrieves all fonts matching the given family or font name."""
    if __FONTCACHE is None:
        init()
    if len(__FONTCACHE) == 0:
        return None

    results = []
    rappend = results.append

    name = name.lower()
    if ftype:
        ftype = ftype.lower()

    fonts = __FONTCACHE.get(name, [])
    for fname, fstyles, fonttype, filename in fonts:
        if ftype and fonttype != ftype:
            # ignore font filetype mismatches
            continue
        if style == STYLE_NONE or fstyles == style:
            rappend((name, fname, fstyles, fonttype, filename))

    for family, fonts in __FONTCACHE.items():
        for fname, fstyles, fonttype, filename in fonts:
            if fname.lower() == name and filename not in results:
                rappend((family, fname, fstyles, fonttype, filename))
    return results


def get_font(name, style=STYLE_NONE, ftype=None):
    """Retrieves the best matching font file for the given name and
    criteria.
    """
    retvals = get_fonts(name, style, ftype)
    if len(retvals) > 0:
        return retvals[0]
    return None



import sounddevice as sd
import audioop
import time

from .helpers import Config, error_print

sample_width = Config.sample_width

def sample_width_to_dtype():
    if sample_width == 1 or sample_width == 2 or sample_width == 4:
        return 'int' + str(8*sample_width)
    else:
        error_print("Unknown sample width, setting default value 2.")
        Config.sample_width = 2
        return 'int16'

def _latency(latency):
    if latency == "high" or latency == "low":
        return latency
    else:
        return float(latency)

class Mixer:
    def __init__(self):
        self.stream = sd.RawOutputStream(
                samplerate=Config.frame_rate,
                channels=Config.channels,
                dtype=sample_width_to_dtype(),
                latency=_latency(Config.latency),
                blocksize=Config.blocksize,
                callback=self.play_callback)
        self.open_files = []

    def add_file(self, music_file):
        if music_file not in self.open_files:
            self.open_files.append(music_file)
        self.start()

    def remove_file(self, music_file):
        if music_file in self.open_files:
            self.open_files.remove(music_file)
        if len(self.open_files) == 0:
            self.stop()

    def stop(self):
        self.stream.stop()

    def start(self):
        self.stream.start()

    def play_callback(self, out_data, frame_count, time_info, status_flags):
        out_data_length = len(out_data)
        empty_data = b"\0" * out_data_length
        data = b"\0" * out_data_length

        for open_file in self.open_files:
            file_data = open_file.play_callback(out_data_length, frame_count)

            if data == empty_data:
                data = file_data
            elif file_data != empty_data:
                data = audioop.add(data, file_data, sample_width)

        out_data[:] = data




from kivy.uix.relativelayout import RelativeLayout
from kivy.properties import NumericProperty, ListProperty, StringProperty
from kivy.core.window import Window
from kivy.clock import Clock

import threading
import yaml
import sys
from collections import defaultdict

from transitions.extensions import HierarchicalMachine as Machine

from .music_file import MusicFile
from .mixer import Mixer
from .helpers import Config, gain, error_print, warn_print
from .action import Action

class Mapping(RelativeLayout):
    STATES = [
        'initial',
        'configuring',
        'configured',
        'loading',
        'loaded',
        'failed'
    ]

    TRANSITIONS = [
        {
            'trigger': 'configure',
            'source': 'initial',
            'dest': 'configuring'
        },
        {
            'trigger': 'fail',
            'source': 'configuring',
            'dest': 'failed'
        },
        {
            'trigger': 'success',
            'source': 'configuring',
            'dest': 'configured',
            'after': 'load'
        },
        {
            'trigger': 'load',
            'source': 'configured',
            'dest': 'loading'
        },
        {
            'trigger': 'fail',
            'source': 'loading',
            'dest': 'failed'
        },
        {
            'trigger': 'success',
            'source': 'loading',
            'dest': 'loaded'
        },
        {
            'trigger': 'reload',
            'source': 'loaded',
            'dest': 'configuring'
        }
    ]

    master_volume = NumericProperty(100)
    ready_color = ListProperty([1, 165/255, 0, 1])
    state = StringProperty("")

    def __init__(self, **kwargs):
        self.keys = []
        self.running = []
        self.wait_ids = {}
        self.open_files = {}

        Machine(model=self, states=self.STATES,
                transitions=self.TRANSITIONS, initial='initial',
                ignore_invalid_triggers=True, queued=True)
        super(Mapping, self).__init__(**kwargs)
        self.keyboard = Window.request_keyboard(self.on_keyboard_closed, self)
        self.keyboard.bind(on_key_down=self.on_keyboard_down)

        self.configure()

    def on_enter_configuring(self):
        if Config.builtin_mixing:
            self.mixer = Mixer()
        else:
            self.mixer = None

        try:
            self.key_config, self.open_files = self.parse_config()
        except Exception as e:
            error_print("Error while loading configuration: {}".format(e),
                    with_trace=True, exit=True)
        else:
            self.success()

    def on_enter_loading(self):
        for key in self.keys:
            key.reload()
        self.success()

    # Kivy events
    def add_widget(self, widget, index=0):
        if type(widget).__name__ == "Key" and widget not in self.keys:
            self.keys.append(widget)
        return super(Mapping, self).add_widget(widget, index)

    def remove_widget(self, widget, index=0):
        if type(widget).__name__ == "Key" and widget in self.keys:
            self.keys.remove(widget)
        return super(Mapping, self).remove_widget(widget, index)

    def on_keyboard_closed(self):
        self.keyboard.unbind(on_key_down=self.on_keyboard_down)
        self.keyboard = None

    def on_keyboard_down(self, keyboard, keycode, text, modifiers):
        key = self.find_by_key_code(keycode)
        if self.allowed_modifiers(modifiers) and key is not None:
            modifiers.sort()
            threading.Thread(name="MSKeyAction", target=key.run,
                    args=['-'.join(modifiers)]).start()
        elif 'ctrl' in modifiers and (keycode[0] == 113 or keycode[0] == '99'):
            self.keyboard.unbind(on_key_down=self.on_keyboard_down)
            self.stop_all_running()
            for music in self.open_files.values():
                music.stop()
            for thread in threading.enumerate():
                if thread.getName()[0:2] == "MS":
                    thread.join()
                elif thread.__class__ == threading.Timer:
                    thread.cancel()
                    thread.join()

            sys.exit()
        elif 'ctrl' in modifiers and keycode[0] == 114:
            threading.Thread(name="MSReload", target=self.reload).start()
        return True

    # Helpers
    def allowed_modifiers(self, modifiers):
        allowed = []
        return len([a for a in modifiers if a not in allowed]) == 0

    def find_by_key_code(self, key_code):
        if "Key_" + str(key_code[0]) in self.ids:
            return self.ids["Key_" + str(key_code[0])]
        return None

    def all_keys_ready(self):
        partial = False
        for key in self.keys:
            if not key.is_loaded_or_failed():
                return "not_ready"
            partial = partial or key.is_failed()

        if partial:
            return "partial"
        else:
            return "success"

    # Callbacks
    def key_loaded_callback(self):
        result = self.all_keys_ready()
        if result == "success":
            self.ready_color = [0, 1, 0, 1]
        elif result == "partial":
            self.ready_color = [1, 0, 0, 1]
        else:
            self.ready_color = [1, 165/255, 0, 1]

    ## Some global actions
    def stop_all_running(self, except_key=None, key_start_time=0):
        running = self.running
        self.running = [r for r in running\
                if r[0] == except_key and r[1] == key_start_time]
        for (key, start_time) in running:
            if (key, start_time) != (except_key, key_start_time):
                key.interrupt()

    # Master volume methods
    @property
    def master_gain(self):
        return gain(self.master_volume)

    def set_master_volume(self, value, delta=False, fade=0):
        [db_gain, self.master_volume] = gain(
                value + int(delta) * self.master_volume,
                self.master_volume)

        for music in self.open_files.values():
            music.set_gain_with_effect(db_gain, fade=fade)

    # Wait handler methods
    def add_wait_id(self, wait_id, action_or_wait):
        self.wait_ids[wait_id] = action_or_wait

    def interrupt_wait(self, wait_id):
        if wait_id in self.wait_ids:
            action_or_wait = self.wait_ids[wait_id]
            del(self.wait_ids[wait_id])
            if isinstance(action_or_wait, Action):
                action_or_wait.interrupt()
            else:
                action_or_wait.set()

    # Methods to control running keys
    def start_running(self, key, start_time):
        self.running.append((key, start_time))

    def keep_running(self, key, start_time):
        return (key, start_time) in self.running

    def finished_running(self, key, start_time):
        if (key, start_time) in self.running:
            self.running.remove((key, start_time))

    # YML config parser
    def parse_config(self):
        def update_alias(prop_hash, aliases, key):
            if isinstance(aliases[key], dict):
                prop_hash.update(aliases[key], **prop_hash)
            else:
                warn_print("Alias {} is not a hash, ignored".format(key))

        def include_aliases(prop_hash, aliases):
            if 'include' not in prop_hash:
                return

            included = prop_hash['include']
            del(prop_hash['include'])
            if isinstance(included, str):
                update_alias(prop_hash, aliases, included)
            elif isinstance(included, list):
                for included_ in included:
                    if isinstance(included_, str):
                        update_alias(prop_hash, aliases, included_)
                    else:
                        warn_print("Unkown alias include type, ignored: "
                            "{} in {}".format(included_, included))
            else:
                warn_print("Unkown alias include type, ignored: {}"
                        .format(included))

        def check_key_property(key_property, key):
            if 'description' in key_property:
                desc = key_property['description']
                if not isinstance(desc, list):
                    warn_print("description in key_property '{}' is not "
                            "a list, ignored".format(key))
                    del(key_property['description'])
            if 'color' in key_property:
                color = key_property['color']
                if not isinstance(color, list)\
                        or len(color) != 3\
                        or not all(isinstance(item, int) for item in color)\
                        or any(item < 0 or item > 255 for item in color):
                    warn_print("color in key_property '{}' is not "
                            "a list of 3 valid integers, ignored".format(key))
                    del(key_property['color'])

        def check_key_properties(config):
            if 'key_properties' in config:
                if isinstance(config['key_properties'], dict):
                    return config['key_properties']
                else:
                    warn_print("key_properties config is not a hash, ignored")
                    return {}
            else:
                return {}

        def check_mapped_keys(config):
            if 'keys' in config:
                if isinstance(config['keys'], dict):
                    return config['keys']
                else:
                    warn_print("keys config is not a hash, ignored")
                    return {}
            else:
                return {}

        def check_mapped_key(mapped_keys, key):
            if not isinstance(mapped_keys[key], list):
                warn_print("key config '{}' is not an array, ignored"
                        .format(key))
                return []
            else:
                return mapped_keys[key]

        def check_music_property(music_property, filename):
            if not isinstance(music_property, dict):
                warn_print("music_property config '{}' is not a hash, ignored"
                        .format(filename))
                return {}
            if 'name' in music_property:
                music_property['name'] = str(music_property['name'])
            if 'gain' in music_property:
                try:
                    music_property['gain'] = float(music_property['gain'])
                except ValueError as e:
                    del(music_property['gain'])
                    warn_print("gain for music_property '{}' is not "
                            "a float, ignored".format(filename))
            return music_property

        stream = open(Config.yml_file, "r")
        try:
            config = yaml.safe_load(stream)
        except Exception as e:
            error_print("Error while loading config file: {}".format(e),
                    exit=True)
        stream.close()

        if not isinstance(config, dict):
            error_print("Top level config is supposed to be a hash",
                    exit=True)

        if 'aliases' in config and isinstance(config['aliases'], dict):
            aliases = config['aliases']
        else:
            aliases = defaultdict(dict)
            if 'aliases' in config:
                warn_print("aliases config is not a hash, ignored")

        music_properties = defaultdict(dict)
        if 'music_properties' in config and\
                isinstance(config['music_properties'], dict):
            music_properties.update(config['music_properties'])
        elif 'music_properties' in config:
            warn_print("music_properties config is not a hash, ignored")

        seen_files = {}

        key_properties = defaultdict(lambda: {
                "actions":    [],
                "properties": {},
                "files":      []
            })

        for key in check_key_properties(config):
            key_prop = config['key_properties'][key]

            if not isinstance(key_prop, dict):
                warn_print("key_property '{}' is not a hash, ignored"
                        .format(key))
                continue

            include_aliases(key_prop, aliases)
            check_key_property(key_prop, key)

            key_properties[key]["properties"] = key_prop

        for mapped_key in check_mapped_keys(config):
            for index, action in enumerate(check_mapped_key(
                    config['keys'], mapped_key)):
                if not isinstance(action, dict) or\
                        not len(action) == 1 or\
                        not isinstance(list(action.values())[0] or {}, dict):
                    warn_print("action number {} of key '{}' is invalid, "
                            "ignored".format(index + 1, mapped_key))
                    continue

                action_name = list(action)[0]
                action_args = {}
                if action[action_name] is None:
                    action[action_name] = {}

                include_aliases(action[action_name], aliases)

                for argument in action[action_name]:
                    if argument == 'file':
                        filename = str(action[action_name]['file'])
                        if filename not in seen_files:
                            music_property = check_music_property(
                                    music_properties[filename],
                                    filename)

                            if filename in self.open_files:
                                self.open_files[filename]\
                                        .reload_properties(**music_property)

                                seen_files[filename] =\
                                        self.open_files[filename]
                            else:
                                seen_files[filename] = MusicFile(
                                        filename, self, **music_property)

                        if filename not in key_properties[mapped_key]['files']:
                            key_properties[mapped_key]['files'] \
                                    .append(seen_files[filename])

                        action_args['music'] = seen_files[filename]
                    else:
                        action_args[argument] = action[action_name][argument]

                key_properties[mapped_key]['actions'] \
                        .append([action_name, action_args])

        return (key_properties, seen_files)





from .helpers import parse_args, register_fonts, path

parse_args()

import kivy
kivy.require("1.9.1")
from kivy.app import App
from kivy.uix.floatlayout import FloatLayout
from kivy.uix.relativelayout import RelativeLayout
from kivy.uix.label import Label
from kivy.properties import ListProperty, StringProperty
from kivy.core.window import Window
from kivy.lang import Builder

register_fonts()


from .helpers import Config
from .key import Key
from .mapping import Mapping

from .app_blocks.actionlist import *
from .app_blocks.playlist import *

class KeyList(RelativeLayout):
    keylist = ListProperty([])
    first_key = StringProperty("")
    second_key = StringProperty("")
    third_key = StringProperty("")

    def append(self, value):
        self.keylist.insert(0, value)

    def on_keylist(self, instance, new_key_list):
        if len(self.keylist) > 0:
            self.first_key  = self.keylist[0]
        if len(self.keylist) > 1:
            self.second_key = self.keylist[1]
        if len(self.keylist) > 2:
            self.third_key  = self.keylist[2]

class UnfocusedOverlay(Label):
    pass

class Screen(FloatLayout):
    def __init__(self, **kwargs):
        super(Screen, self).__init__(**kwargs)
        self.unfocused_widget = UnfocusedOverlay()
        Window.bind(focus=self.focus_changed)

    def focus_changed(self, instance, focus):
        if Config.no_focus_warning:
            return
        if not focus:
            self.add_widget(self.unfocused_widget)
        else:
            self.remove_widget(self.unfocused_widget)

class MusicSamplerApp(App):
    def build(self):
        Window.size = (913, 563)

        return Screen()

def main():
    Builder.load_file(path() + "/music_sampler.kv")
    MusicSamplerApp().run()



from kivy.uix.label import Label
from kivy.uix.stacklayout import StackLayout
from kivy.uix.relativelayout import RelativeLayout
from kivy.properties import ListProperty, StringProperty
from ..lock import Lock

__all__ = ["ActionList",
        "ActionListIcons", "ActionListIcon",
        "ActionListDescriptions", "ActionListDescription"]

actionlist_lock = Lock("playlist")

class ActionList(RelativeLayout):
    action_title = StringProperty("")
    action_list = ListProperty([])

    def update_list(self, key, action_descriptions):
        self.action_title = _("actions linked to key {}:").format(key.key_sym)
        action_list = []

        for [action, status] in action_descriptions:
            if status == "done":
                icon = "✓"
            elif status == "current":
                icon = "✅"
            else:
                icon = " "
            action_list.append([icon, action])
        with actionlist_lock:
            self.action_list = action_list

class ActionListIcons(StackLayout):
    def __init__(self, **kwargs):
        super(ActionListIcons, self).__init__(**kwargs)
        self.icons = []

    def on_parent(self, instance, parent):
        parent.bind(action_list=self.update_actionlist_icons)

    def update_actionlist_icons(self, instance, actionlist):
        icons_length = len(self.icons)
        index = -1
        for index, [icon, description] in enumerate(actionlist):
            if index >= icons_length:
                icon_label = ActionListIcon(text=icon)
                self.add_widget(icon_label)
                self.icons.append(icon_label)
            else:
                self.icons[index].text = icon

        if index+1 < icons_length:
            self.clear_widgets(children=self.icons[index+1:icons_length])
            del(self.icons[index+1:icons_length])

class ActionListIcon(Label):
    def __init__(self, text='', **kwargs):
        super(ActionListIcon, self).__init__(**kwargs)
        self.text = text

    def on_parent(self, instance, parent):
        if parent is not None:
            parent.bind(font_size=self.update_font_size)
            parent.bind(labels_height=self.update_height)

    def update_height(self, instance, height):
        self.height = height

    def update_font_size(self, instance, font_size):
        self.font_size = font_size

class ActionListDescriptions(StackLayout):
    def __init__(self, **kwargs):
        super(ActionListDescriptions, self).__init__(**kwargs)
        self.descriptions = []

    def on_parent(self, instance, parent):
        parent.bind(action_list=self.update_actionlist_descriptions)

    def update_actionlist_descriptions(self, instance, actionlist):
        descriptions_length = len(self.descriptions)
        index = -1
        for index, [icon, description] in enumerate(actionlist):
            if index >= descriptions_length:
                description_label = ActionListDescription(text=description)
                self.add_widget(description_label)
                self.descriptions.append(description_label)
            else:
                self.descriptions[index].text = description

        if index+1 < descriptions_length:
            self.clear_widgets(
                    children=self.descriptions[index+1:descriptions_length])
            del(self.descriptions[index+1:descriptions_length])

class ActionListDescription(Label):
    def __init__(self, text='', **kwargs):
        super(ActionListDescription, self).__init__(**kwargs)
        self.text = text

    def on_parent(self, instance, parent):
        if parent is not None:
            parent.bind(font_size=self.update_font_size)
            parent.bind(labels_height=self.update_height)

    def update_height(self, instance, height):
        self.height = height

    def update_font_size(self, instance, font_size):
        self.font_size = font_size




import threading

def run(action, duration=0, music=None, set_wait_id=None, **kwargs):
    if set_wait_id is not None:
        action.mapping.add_wait_id(set_wait_id, action)

    action.sleep_event = threading.Event()
    action.sleep_event_timer = threading.Timer(
            duration,
            action.sleep_event.set)

    if music is not None:
        music.wait_end()

    action.sleep_event_timer.start()
    action.sleep_event.wait()

def description(action, duration=0, music=None, set_wait_id=None, **kwargs):
    formats = []
    message = ""
    if music is None:
        message += "waiting {}s"
        formats.append(duration)
    elif duration == 0:
        message += "waiting the end of « {} »"
        formats.append(music.name)
    else:
        message += "waiting the end of « {} » + {}s"
        formats.append(music.name)
        formats.append(duration)

    if set_wait_id is not None:
        message += " (setting id = {})"
        formats.append(set_wait_id)

    return _(message).format(*formats)

def interrupt(action, duration=0, music=None, **kwargs):
    if action.sleep_event is not None:
        action.sleep_event.set()
        action.sleep_event_timer.cancel()
    if music is not None:
        music.wait_event.set()



# -*- coding: utf-8 -*-
import argparse
import sys
import os
import math
import sounddevice as sd
import logging
import gettext
gettext.install('music_sampler')
Logger = logging.getLogger("kivy")

from . import sysfont

class Config:
    pass

def find_font(name, style=sysfont.STYLE_NONE):
    if getattr(sys, 'frozen', False):
        font = sys._MEIPASS + "/fonts/{}_{}.ttf".format(name, style)
    else:
        font = sysfont.get_font(name, style=style)
        if font is not None:
            font = font[4]
    return font

def register_fonts():
    from kivy.core.text import LabelBase

    ubuntu_regular = find_font("Ubuntu", style=sysfont.STYLE_NORMAL)
    ubuntu_bold = find_font("Ubuntu", style=sysfont.STYLE_BOLD)
    symbola = find_font("Symbola")

    if ubuntu_regular is None:
        error_print("Font Ubuntu regular could not be found, "
                "please install it.", exit=True)
    if symbola is None:
        error_print("Font Symbola could not be found, please install it.",
                exit=True)
    if ubuntu_bold is None:
        warn_print("Font Ubuntu Bold could not be found.")

    LabelBase.register(name="Ubuntu",
            fn_regular=ubuntu_regular,
            fn_bold=ubuntu_bold)
    LabelBase.register(name="Symbola",
            fn_regular=symbola)


def path():
    if getattr(sys, 'frozen', False):
        return sys._MEIPASS + "/"
    else:
        return os.path.dirname(os.path.realpath(__file__))

def parse_args():
    argv = sys.argv[1 :]
    sys.argv = sys.argv[: 1]
    if "--" in argv:
        index = argv.index("--")
        kivy_args = argv[index+1 :]
        argv = argv[: index]

        sys.argv.extend(kivy_args)

    os.environ["KIVY_NO_CONFIG"] = 'true'
    sys.argv.extend(["-c", "kivy:log_level:warning"])
    sys.argv.extend(["-c", "kivy:log_dir:/tmp"])
    sys.argv.extend(["-c", "kivy:log_name:/tmp/music_sampler_%_.txt"])

    parser = argparse.ArgumentParser(
            description=_("A Music Sampler application."),
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("-c", "--config",
            default="config.yml",
            required=False,
            help=_("Config file to load"))
    parser.add_argument("-p", "--music-path",
            default=".",
            required=False,
            help=_("Folder in which to find the music files"))
    parser.add_argument("-d", "--debug",
            nargs=0,
            action=DebugModeAction,
            help=_("Print messages in console"))
    parser.add_argument("-m", "--builtin-mixing",
            action="store_true",
            help=_("Make the mixing of sounds manually\
                    (do it if the system cannot handle it correctly)"))
    parser.add_argument("-l", "--latency",
            default="high",
            required=False,
            help=_("Latency: low, high or number of seconds"))
    parser.add_argument("-b", "--blocksize",
            default=0,
            type=int,
            required=False,
            help=_("Blocksize: If not 0, the number of frames to take\
                    at each step for the mixer"))
    parser.add_argument("-f", "--frame-rate",
            default=44100,
            type=int,
            required=False,
            help=_("Frame rate to play the musics"))
    parser.add_argument("-x", "--channels",
            default=2,
            type=int,
            required=False,
            help=_("Number of channels to use"))
    parser.add_argument("-s", "--sample-width",
            default=2,
            type=int,
            required=False,
            help=_("Sample width (number of bytes for each frame)"))
    parser.add_argument("-V", "--version",
            action="version",
            help=_("Displays the current version and exits. Only use\
                    in bundled package"),
            version=show_version())
    parser.add_argument("--device",
            action=SelectDeviceAction,
            help=_("Select this sound device")
            )
    parser.add_argument("--list-devices",
            nargs=0,
            action=ListDevicesAction,
            help=_("List available sound devices")
            )
    parser.add_argument("--no-focus-warning",
            action='store_true',
            help=_("Don't show warning when focus is lost")
            )
    parser.add_argument("-L", "--language",
            required=False,
            default="fr",
            help=_("Select another language")
            )
    parser.add_argument('--',
            dest="args",
            help=_("Kivy arguments. All arguments after this are interpreted\
                    by Kivy. Pass \"-- --help\" to get Kivy's usage."))

    args = parser.parse_args(argv)

    Config.yml_file = args.config

    Config.latency = args.latency
    Config.blocksize = args.blocksize
    Config.frame_rate = args.frame_rate
    Config.channels = args.channels
    Config.sample_width = args.sample_width
    Config.builtin_mixing = args.builtin_mixing
    Config.no_focus_warning = args.no_focus_warning
    if args.language != 'en':
        gettext.translation("music_sampler",
                localedir=path() + '/locales',
                languages=[args.language]).install()
    if args.music_path.endswith("/"):
        Config.music_path = args.music_path
    else:
        Config.music_path = args.music_path + "/"

class DebugModeAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        sys.argv.extend(["-c", "kivy:log_level:debug"])

class SelectDeviceAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        sd.default.device = values

class ListDevicesAction(argparse.Action):
    nargs = 0
    def __call__(self, parser, namespace, values, option_string=None):
        print(sd.query_devices())
        sys.exit()

def show_version():
    if getattr(sys, 'frozen', False):
        with open(path() + ".pyinstaller_commit", "r") as f:
            return f.read()
    else:
        return _("option '-V' can only be used in bundled package")

def duration_to_min_sec(duration):
    minutes = int(duration / 60)
    seconds = int(duration) % 60
    if minutes < 100:
        return "{:2}:{:0>2}".format(minutes, seconds)
    else:
        return "{}:{:0>2}".format(minutes, seconds)

def gain(volume, old_volume=None):
    if old_volume is None:
        return 20 * math.log10(max(volume, 0.1) / 100)
    else:
        return [
                20 * math.log10(max(volume, 0.1) / max(old_volume, 0.1)),
                max(volume, 0)]

def debug_print(message, with_trace=None):
    if with_trace is None:
        with_trace = (Logger.getEffectiveLevel() < logging.WARN)
    with_trace &= (sys.exc_info()[0] is not None)

    Logger.debug('MusicSampler: ' + message, exc_info=with_trace)

def error_print(message, exit=False, with_trace=None):
    if with_trace is None:
        with_trace = (Logger.getEffectiveLevel() < logging.WARN)
    with_trace &= (sys.exc_info()[0] is not None)

    # FIXME: handle it correctly when in a thread
    if exit:
        Logger.critical('MusicSampler: ' + message, exc_info=with_trace)
        sys.exit(1)
    else:
        Logger.error('MusicSampler: ' + message, exc_info=with_trace)

def warn_print(message, with_trace=None):
    if with_trace is None:
        with_trace = (Logger.getEffectiveLevel() < logging.WARN)
    with_trace &= (sys.exc_info()[0] is not None)

    Logger.warn('MusicSampler: ' + message, exc_info=with_trace)




def run(action, music=None, **kwargs):
    for music in action.music_list(music):
        if music.is_loaded_paused():
            music.unpause()

def description(action, music=None, **kwargs):
    if music is not None:
        return _("unpausing « {} »").format(music.name)
    else:
        return _("unpausing all musics")
